package servlet;
import java.io.DataInputStream;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.Socket;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.util.Vector;

import javax.swing.JOptionPane;

import common.ContactList;
import common.Message;
import common.MessageServer;
import common.User;
import common.interfaces.BSForServletInterface;
import common.interfaces.MSForServletInterface;
import common.interfaces.exceptions.AssignationNotAcceptedException;
import common.interfaces.exceptions.GenericErrorException;
import common.interfaces.exceptions.NotAllowedException;
import common.interfaces.exceptions.SystemErrorException;
import common.interfaces.exceptions.UserAlreadyInsertedException;
import common.interfaces.exceptions.UserBlockedException;
import common.interfaces.exceptions.UserNotExistsException;
import common.interfaces.exceptions.UserNotManagedException;
import common.interfaces.exceptions.WrongPasswordException;

/**
 * Provides a whole of methods which to communicate with the server.
 * @author Alessandro Franchini
 *
 */
public class AccessOperation {
	/** The reference to the remote object of the bootstrap	 */
	private static BSForServletInterface boot;
	/** The reference to the remote object of the message	 */
	private static MSForServletInterface msi;
	
	/**
	 * This method finds the reference to the remote object of the bootstrap.
	 * @param bootIa the bootstrap's address
	 * @param port the bootstrap's port
	 * @return true if the object is found; else false.
	 */
	public static boolean setConnBoot(InetAddress bootIa, int port){
		boolean flag = true;
		String Url  = "rmi://"+bootIa.getHostAddress()+":"+port+"/gossipBS";
		try {
			boot = (BSForServletInterface) Naming.lookup(Url);
		} catch (MalformedURLException e) {
			catchError("Error in the name of URL.");
			flag = false;
		} catch (RemoteException e) {
			catchError("A communication problems occurs.");
			flag = false;
		} catch (NotBoundException e) {
			catchError("Could not find the requested remote object on the server");
		}
		return flag;
	}
	
	/**
	 * This method is of use to the user's registration
	 * @param user the user
	 * @param pass the user's pass
	 * @return true if the registration has been made; else false
	 */
	public static boolean register(User user, String pass){
		boolean flag = true;
		try {	
			boot.register(user.getNick(), pass);
		} catch (RemoteException e) {
			catchError("A communication problems occurs.");
			flag = false;
		} catch (GenericErrorException e) {
			catchError("There are no message server available.");
			flag = false;
		} catch (AssignationNotAcceptedException e) {
			catchError("Another user with the same nickname already exists.");
			flag = false;
		}
		return flag;
	}
	
	/**
	 * Asks the assignation of a Message Server
	 * @param user the user who to assign the Message Server to.
	 * @return The Message Server assigned
	 */
	public static MessageServer askMsgServer(User user){
		MessageServer msg = null;
		try {
			msg = boot.getMS(user);
		} catch (RemoteException e) {
			catchError("A communication problems occurs.");
		} catch (UserNotExistsException e) {
			catchError("The user does not exist.");
		}
		 
		return msg;
	}
	
	/**
	 * This method finds the reference to the remote object of the message server,
	 * sets the State to true and downloads the Contacts List
	 * @param user the user
	 * @param pass the user's password
	 * @param ms the message server
	 * @return the contact list.
	 */
	public static ContactList logMsg(User user, String pass, MessageServer ms){
		ContactList list = null;
		String Url = "rmi://"+ms.getIP().getHostAddress()+":"+ms.getPort()+"/gossipMS";
		try {
			msi = (MSForServletInterface) Naming.lookup(Url);
			msi.setState(user, pass, true);
			list = msi.getContactList(user, pass);
		} catch (RemoteException e){
			catchError("A communication problems occurs. with MS");
		} catch (MalformedURLException e) {
			catchError("Error in the name of URL.");
		} catch (NotBoundException e) {
			catchError("Could not find the requested remote object on the server");
		} catch (UserNotManagedException e) {
			catchError("The user is not managed by this message server.");
		} catch (WrongPasswordException e) {
			catchError("The password is wrong");
		} catch (GenericErrorException e) {
			catchError("An error occured");
		} catch (UserBlockedException e) {
			catchError("The user is blocked");
		}
		return list;
		
	}
	
	/**
	 * Downloads and deletes the messages received by the user
	 * @param user the user
	 * @param pass the user's password
	 * @param ms the message server
	 * @return the message received by the user
	 */
	public static Vector <Message> getMail(User user, String pass, MessageServer ms){
		Vector <Message> mail = null;
		try{
			mail = msi.downloadMsgs(user, pass);
			msi.deleteDownloadedMsgs(user, pass);
		} catch (RemoteException e) {
			catchError("A communication problems occurs.");
		} catch (UserNotManagedException e) {
			catchError("The user is not managed by this message server.");
			try {
				ClientCore.ms = boot.getMS(ClientCore.myself);
				String Url = "rmi://"+ClientCore.ms.getIP().getHostAddress()+":"+ClientCore.ms.getPort()+"/gossipMS";
				msi = (MSForServletInterface) Naming.lookup(Url);
				msi.setState(user, pass, true);
			} catch (RemoteException e1) {
				catchError("A communication problems occurs.");
			} catch (UserNotExistsException e1) {
				catchError("The user does not exist.");
			} catch (MalformedURLException e1) {
				catchError("Error in the name of URL.");
			} catch (NotBoundException e1) {
				catchError("Could not find the requested remote object on the server");
			} catch (UserNotManagedException e1) {
				catchError("The user is not managed by this message server.");
			} catch (WrongPasswordException e1) {
				catchError("The password is wrong");
			} catch (GenericErrorException e1) {
				catchError("An error occured");
			} catch (UserBlockedException e1) {
				catchError("The user is blocked");
			}
		} catch (UserBlockedException e) {
			catchError("The user is blocked");
		} catch (WrongPasswordException e) {
			catchError("The password is wrong");
		}
		return mail;
	}
	
	/**
	 * Changes the friendship's state with the user
	 * @param form 0 to unlock the contact; 1 to lock the contact
	 * @param user the user
	 */
	public static void changeStateUser(int form, User user) {
		if(form == 0){
			try {
				msi.addToFriends(ClientCore.myself,ClientCore.pwd,user);
			} catch (RemoteException e) {
				catchError("A communication problems occurs.");
			} catch (UserNotManagedException e) {
				catchError("The user is not managed by this message server.");
				try {
					ClientCore.ms = boot.getMS(ClientCore.myself);
					String Url = "rmi://"+ClientCore.ms.getIP().getHostAddress()+":"+ClientCore.ms.getPort()+"/gossipMS";
					msi = (MSForServletInterface) Naming.lookup(Url);
					msi.setState(user, ClientCore.pwd, true);
					msi.addToFriends(ClientCore.myself,ClientCore.pwd,user);
				} catch (RemoteException e1) {
					catchError("A communication problems occurs.");
				} catch (UserNotExistsException e1) {
					catchError("The user does not exist.");
				} catch (MalformedURLException e1) {
					catchError("Error in the name of URL.");
				} catch (NotBoundException e1) {
					catchError("Could not find the requested remote object on the server");
				} catch (UserNotManagedException e1) {
					catchError("The user is not managed by this message server.");
				} catch (WrongPasswordException e1) {
					catchError("The password is wrong");
				} catch (GenericErrorException e1) {
					catchError("An error occured");
				} catch (UserBlockedException e1) {
					catchError("The user is blocked");
				} catch (UserAlreadyInsertedException e1) {
					catchError("The user is already inserted.");
				}
			} catch (WrongPasswordException e) {
				catchError("The password is wrong");
			} catch (UserBlockedException e) {
				catchError("The user is blocked");
			} catch (UserAlreadyInsertedException e) {
				catchError("The user is already inserted.");
			}
		}
		else{
			try{
				msi.blockContact(ClientCore.myself,ClientCore.pwd,user);
			} catch (UserNotManagedException e) {
				catchError("The user is not managed by this message server.");
				try {
					ClientCore.ms = boot.getMS(ClientCore.myself);
					String Url = "rmi://"+ClientCore.ms.getIP().getHostAddress()+":"+ClientCore.ms.getPort()+"/gossipMS";
					msi = (MSForServletInterface) Naming.lookup(Url);
					msi.setState(user, ClientCore.pwd, true);
					msi.blockContact(ClientCore.myself,ClientCore.pwd,user);
				} catch (RemoteException e1) {
					catchError("A communication problems occurs.");
				} catch (UserNotExistsException e1) {
					catchError("The user does not exist.");
				} catch (MalformedURLException e1) {
					catchError("Error in the name of URL.");
				} catch (NotBoundException e1) {
					catchError("Could not find the requested remote object on the server");
				} catch (UserNotManagedException e1) {
					catchError("The user is not managed by this message server.");
				} catch (WrongPasswordException e1) {
					catchError("The password is wrong");
				} catch (GenericErrorException e1) {
					catchError("An error occured");
				} catch (UserBlockedException e1) {
					catchError("The user is blocked");
				}
			} catch (WrongPasswordException e) {
				catchError("The password is wrong");
			} catch (UserNotExistsException e) {
				catchError("The user does not exists");
			} catch (RemoteException e) {
				catchError("A communication problems occurs.");
			} catch (UserBlockedException e) {
				catchError("The user is blocked");
			}
		}	
	}
	
	/**
	 * Set the user's state to false.
	 */
	public static void logOff(){
		try {
			msi.setState(ClientCore.myself,ClientCore.pwd,false);
		} catch (RemoteException e) {
			catchError("A communication problems occurs.");
		} catch (UserNotManagedException e) {
			catchError("The user is not managed by this message server.");
			try {
				ClientCore.ms = boot.getMS(ClientCore.myself);
				String Url = "rmi://"+ClientCore.ms.getIP().getHostAddress()+":"+ClientCore.ms.getPort()+"/gossipMS";
				msi = (MSForServletInterface) Naming.lookup(Url);
				msi.setState(ClientCore.myself, ClientCore.pwd, true);
				msi.setState(ClientCore.myself,ClientCore.pwd,false);
			} catch (RemoteException e1) {
				catchError("A communication problems occurs.");
			} catch (UserNotExistsException e1) {
				catchError("The user does not exist.");
			} catch (MalformedURLException e1) {
				catchError("Error in the name of URL.");
			} catch (NotBoundException e1) {
				catchError("Could not find the requested remote object on the server");
			} catch (UserNotManagedException e1) {
				catchError("The user is not managed by this message server.");
			} catch (WrongPasswordException e1) {
				catchError("The password is wrong");
			} catch (GenericErrorException e1) {
				catchError("An error occured");
			} catch (UserBlockedException e1) {
				catchError("The user is blocked");
			}
		} catch (WrongPasswordException e) {
			catchError("The password is wrong");
		} catch (GenericErrorException e) {
			catchError("An error occured");
		} catch (UserBlockedException e) {
			catchError("The user is blocked");
		}
	}
	
	/**
	 * Sends a message to the message server
	 * @param mx the message server
	 */
	public static void sendMessage(Message mx){
		try {
			msi.forwardMessage(mx);
		} catch (RemoteException e) {
			catchError("A communication problems occurs.");
		} catch (NotAllowedException e) {
			catchError("Not allowed error");
		} catch (UserNotManagedException e) {
			catchError("The user is not managed by this message server.");
			try {
				ClientCore.ms = boot.getMS(ClientCore.myself);
				String Url = "rmi://"+ClientCore.ms.getIP().getHostAddress()+":"+ClientCore.ms.getPort()+"/gossipMS";
				msi = (MSForServletInterface) Naming.lookup(Url);
				msi.setState(ClientCore.myself, ClientCore.pwd, true);
				msi.forwardMessage(mx);
			} catch (RemoteException e1) {
				catchError("A communication problems occurs.");
			} catch (UserNotExistsException e1) {
				catchError("The user does not exist.");
			} catch (MalformedURLException e1) {
				catchError("Error in the name of URL.");
			} catch (NotBoundException e1) {
				catchError("Could not find the requested remote object on the server");
			} catch (UserNotManagedException e1) {
				catchError("The user is not managed by this message server.");
			} catch (WrongPasswordException e1) {
				catchError("The password is wrong");
			} catch (GenericErrorException e1) {
				catchError("An error occured");
			} catch (UserBlockedException e1) {
				catchError("The user is blocked");
			} catch (NotAllowedException e1) {
				catchError("Not allowed error");
			} catch (SystemErrorException e1) {
				catchError("A system error occured");
			}
		} catch (GenericErrorException e) {
			catchError("An error occured");
		} catch (UserNotExistsException e) {
			catchError("The user does not exists");
		} catch (SystemErrorException e) {
			catchError("A system error occured");
		}
	}
	
	/**
	 * Gets information about the requested user
	 * @param user the user
	 * @return The user with all the information
	 */
	public static User getUser(User user){
		User temp = null;
		try {
			temp= msi.getUserContact(user);
		} catch (RemoteException e) {
			catchError("A communication problems occurs.");
		} catch (UserNotExistsException e) {
			catchError("The user does not exists");
		} catch (GenericErrorException e) {
			catchError("A problem is occurred");
		} catch (UserBlockedException e) {
			catchError("The user is blocked");
		}
		return temp;
	}
	
	/**
	 * Lets contact another user the first time
	 * @param user the user
	 * @return true if the user is online; else false.
	 */
	public static boolean contactUser(User user){
		boolean flag = false;
		Message mx = new Message("Type::Contact",user,ClientCore.myself);
		try {
			Socket so = new Socket(user.getIP(),user.getPort());
			OutputStream os = so.getOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream(os);
			oos.writeObject(mx);
			InputStream is = so.getInputStream();
			DataInputStream dis = new DataInputStream(is);
			flag = dis.readBoolean();
			so.close();
		} catch (Exception e) {
			AccessOperation.sendMessage(mx);
			flag= false;
		}
		return flag;
	}
	
	/**
	 * Visualizes a graphic interface with the error
	 * @param error the error
	 */
	private static void catchError(String error){
		JOptionPane.showMessageDialog(null, error, "Gossip", JOptionPane.ERROR_MESSAGE);
		
	}
}
